# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['malloovia']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'jsonschema>=4.17.3,<5.0.0',
 'progress>=1.6,<2.0',
 'pulp==2.0',
 'ruamel.yaml>=0.17.21,<0.18.0']

setup_kwargs = {
    'name': 'malloovia',
    'version': '2.3.2',
    'description': 'Use linear programming to allocate applications to cloud infrastructure',
    'long_description': '# Malloovia\n\n[<img src="https://img.shields.io/badge/python->=3.6-blue.svg?style=flat-square">](https://www.python.org/downloads/) [<img src="https://img.shields.io/pypi/v/malloovia.svg">](https://pypi.python.org/pypi/malloovia) [<img src="http://readthedocs.org/projects/malloovia/badge/?version=latest" alt="Documentation Status">](http://malloovia.readthedocs.io/en/latest/?badge=latest)  [<img src="https://travis-ci.org/asi-uniovi/malloovia.svg?branch=master" alt=" Build status">](https://travis-ci.org/asi-uniovi/malloovia)\n\n\nUse linear programming to allocate applications to cloud infrastructure.\n\n\n* Free software: MIT license\n* Documentation: https://malloovia.readthedocs.io.\n\n\n## Introduction\n\nMalloovia is a Python package to solve virtual machine (VM) allocation problems in Infrastructure as a Service (IaaS) clouds from the point of view of the cloud customer. It was first presented in the paper ["Cost Minimization of Virtual Machine Allocation in Public Clouds Considering Multiple Applications"](http://www.atc.uniovi.es/personal/joaquin-entrialgo/pdfs/Entrialgo2017-gecon.pdf) presented at [GECON 2017](http://2017.gecon-conference.org/).\n\nThe problem to solve is: given a cloud infrastructure composed of different virtual machine types, each one with its own hardware characteristics, and prices, some of them with different pricing schemas, such as discounts for reservation over long periods, and given a set of applications to run on that infrastructure, each one with a different performance on each different VM type, and with a different workload over time, find the number of VMs of each type to activate at each timeslot for each application, so that the expected workload is fulfilled for all applications, the cloud provider limits are not exceeded and the total cost is minimized.\n\nIt works in two phases: first, it computes the number of reserved VMs using a Long Term Workload Prediction (LTWP) and then, it computes the number of on-demand for each time slot using a Short Term Workload Prediction (STWP).\n\nMalloovia can be directly used in Python or by a [CLI interface](https://malloovia.readthedocs.io/en/latest/cli.html#cli). The problems and the solutions can be saved using a [YAML format](https://malloovia.readthedocs.io/en/latest/yaml.html).\n\nThis is an example that assumes that the problem definition is in `problems.yaml`, with `problem1` describing the LTWP and `problem2` describing the STWP:\n\n```\n$ malloovia solve problems.yaml --phase-i-id=problem1 --phase-ii-id=problem2\nReading problems.yaml...(0.004s)\nSolving phase I...(0.020s)\nSolving Phase II |███████████████████████████████████| 100.0% - ETA: 0:00:00\n(0.101s)\nWriting solutions in problems-sol.yaml...(0.006s)\n```\n\nThis is an example in Python (explained in more detail in the [Usage section of the documentation](https://malloovia.readthedocs.io/en/latest/usage.html)):\n\n```python\nfrom malloovia import *\n\n# Infrastructure definition\nregion1 = LimitingSet("r1", name="us.east", max_vms=20)\nzone1 =  LimitingSet("r1_z1", name="us.east_a", max_vms=20)\nm3large_z1 = InstanceClass(\n    "m3large_r1_z1", name="reserved m3.large in us.east_a",\n    limiting_sets=(zone1,), is_reserved=True,\n    price=7, time_unit="h", max_vms=20)\nm4xlarge_r1 = InstanceClass(\n    "m4xlarge_r1", name="ondemand m4.xlarge in us.east",\n    limiting_sets=(region1,), is_reserved=False,\n    price=10, time_unit="h", max_vms=10)\n\n# Performances\napp0 = App("a0", "Web server")\napp1 = App("a1", "Database")\nperformances = PerformanceSet(\n    id="example_perfs",\n    time_unit="h",\n    values=PerformanceValues({\n        m3large_z1: {app0: 12, app1: 500},\n        m4xlarge_r1: {app0: 44, app1: 1800}\n        })\n)\n\n# Workload\n\n# Long term workload prediction of each app, for Phase I\nltwp_app0 = Workload(\n    "ltwp0", description="rph to the web server", app=app0,\n    values=(201, 203, 180, 220, 190, 211, 199, 204, 500, 200)\n)\nltwp_app1 = Workload(\n    "ltwp1", description="rph to the database", app=app1,\n    values=(2010, 2035, 1807, 2202, 1910, 2110, 1985, 2033, 5050, 1992)\n)\n\n# Building the problem for phase I and solving\nproblem = Problem(\n    id="example1",\n    name="Example problem",\n    workloads=(ltwp_app0, ltwp_app1),\n    instance_classes=(m3large_z1, m4xlarge_r1),\n    performances=performances\n)\n\nphase_i_solution = PhaseI(problem).solve()\n\n# Building the problem for a timeslot in phase II and solving\nphase_ii = PhaseII(problem, phase_i_solution)\ntimeslot_solution = phase_ii.solve_timeslot(\n    workloads=(Workload("stwp0", app=app0, description=None, values=(315,)),\n               Workload("stwp1", app=app1, description=None, values=(1950,))\n               )\n    )\n    \n# Showing the cost and the allocation\nprint("Cost:", timeslot_solution.solving_stats.optimal_cost)\nprint(timeslot_solution.allocation._inspect())\n```\n\nYou can find example problems and solutions in YAML format in the [test data directory](https://github.com/asi-uniovi/malloovia/tree/master/tests/test_data/valid) and in the [GECON 2017 data repository](https://github.com/asi-uniovi/malloovia-data-gecon2017), where you can find [a notebook](https://github.com/asi-uniovi/malloovia-data-gecon2017/blob/master/Malloovia-Gecon2017-data.ipynb) that shows how to compute the solutions from the problems.\n\nPlease, refer to [the documentation](https://malloovia.readthedocs.io/) and the he paper ["Cost Minimization of Virtual Machine Allocation in Public Clouds Considering Multiple Applications"](http://www.atc.uniovi.es/personal/joaquin-entrialgo/pdfs/Entrialgo2017-gecon.pdf) for more details.\n',
    'author': 'Jose Luis Diaz',
    'author_email': 'jldiaz@uniovi.es',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
